# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################

"""Complex settings at runtime
===========================

Allow to define settings based on references to other settings (overriden in other config files).
Some examples are given below.

.. code-block:: python

  # file1: df_config.config.defaults
  from df_config.config.dynamic_settings import SettingReference
  DEBUG = False
  TEMPLATE_DEBUG = SettingReference('DEBUG')


.. code-block:: python

  # file1: myproject.defaults
  DEBUG = True


Since the second file overrides the first one, `TEMPLATE_DEBUG` always has the same value as `DEBUG` (`True`).

"""
import os
import socket
import sys
from typing import Any, Set
from urllib.parse import urlparse

from django.core.checks import Warning
from django.utils.module_loading import import_string

from df_config.checks import settings_check_results


class DynamicSettting:
    """Base class for special setting values. When a setting is a
    :class:`df_config.config.dynamic_settings.DynamicSetting`,
      then the method `get_value(merger)` is called for getting the definitive value.
    """

    def __init__(self, value: Any):
        self.value = value

    def get_value(self, merger, provider_name: str, setting_name: str):
        """Return the intepreted value.

        :param merger: merger object, with all interpreted settings
        :type merger: :class:`df_config.config.merger.SettingMerger`
        :param provider_name: name of the provider containing this value
        :param setting_name: name of the setting containing this value
        """
        raise NotImplementedError

    # noinspection PyMethodMayBeStatic
    def pre_collectstatic(self, merger, provider_name, setting_name, value):
        """Add extra commands before the "collectstatic" command (at least the one provided by Djangofloor).
        Could be used for creating public files or directories (static files, required directories...).
        """
        pass

    # noinspection PyMethodMayBeStatic
    def pre_migrate(self, merger, provider_name, setting_name, value):
        """Add extra commands before the "migrate" command.
        Could be used for creating private files (like the SECRET_KEY file)
        """
        pass

    # noinspection PyMethodMayBeStatic
    def post_collectstatic(self, merger, provider_name, setting_name, value):
        """Add extra commands after the "collectstatic" command"""
        pass

    # noinspection PyMethodMayBeStatic
    def post_migrate(self, merger, provider_name, setting_name, value):
        """Add extra commands  after the "migrate" command"""
        pass

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.value)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and other.value == self.value

    @staticmethod
    def add_warning(msg, excluded_commands: Set[str] = None):
        if (
            excluded_commands
            and sys.argv[1:2]
            and sys.argv[1:2][0] in excluded_commands
        ):
            return
        settings_check_results.append(
            Warning(
                msg,
                obj="configuration",
            )
        )


class RawValue(DynamicSettting):
    """Return the value as-is. Since by defaults all string values are assumed to be formatted string,
    you need to use :class:`RawValue` for using values that should be formatted.

    .. code-block:: python

        from df_config.config.dynamic_settings import RawValue
        SETTING_1 = '{DEBUG}'  # will be transformed to 'True' or 'False'
        SETTING_2 = Raw('{DEBUG}')  # will be kept as  '{DEBUG}'
    """

    def get_value(self, merger, provider_name: str, setting_name: str):
        """Return the non-intepreted value

        :param merger: merger object, with all interpreted settings
        :type merger: :class:`df_config.config.merger.SettingMerger`
        :param provider_name: name of the provider containing this value
        :param setting_name: name of the setting containing this value
        """
        return self.value


class Path(DynamicSettting):
    """Represent any normalized path on the filesystem."""

    def __init__(self, value, mode=None):
        super().__init__(value)
        self.mode = mode

    def get_value(self, merger, provider_name, setting_name):
        """Return the value

        :param merger: merger object, with all interpreted settings
        :type merger: :class:`df_config.config.merger.SettingMerger`
        :param provider_name: name of the provider containing this value
        :param setting_name: name of the setting containing this value
        """
        value = merger.analyze_raw_value(self.value, provider_name, setting_name)
        value = os.path.normpath(value)
        return value

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return "%s('%s')" % (self.__class__.__name__, str(self.value))

    @staticmethod
    def makedirs(merger, dirname):
        if not dirname or os.path.isdir(dirname):
            return
        dirname = os.path.abspath(dirname)
        if os.path.exists(dirname):
            merger.stderr.write("'%s' already exists and is not a directory.")
            return
        merger.stdout.write(f"Creating directory '{dirname}'")
        try:
            os.makedirs(dirname)
        except Exception as e:
            merger.stderr.write(f"Unable to create directory '{dirname}' ({e})")

    def chmod(self, merger, filename):
        if not filename or not os.path.exists(filename):
            return
        filename = os.path.abspath(filename)
        if self.mode is None or (os.stat(filename).st_mode & 0o777) == self.mode:
            return
        merger.stdout.write("Change mode of '%s' to 0o%o" % (filename, self.mode))
        try:
            os.chmod(filename, self.mode)
        except Exception as e:
            merger.stderr.write(f"Unable to change mode of '{filename}' ({e})")


class Directory(Path):
    """Represent a directory on the filesystem, that is automatically created by the "migrate" and "collectstatic"
    commands. The resulting value always finishes by a '/'.
    """

    def get_value(self, merger, provider_name: str, setting_name: str):
        """Return the value

        :param merger: merger object, with all interpreted settings
        :type merger: :class:`df_config.config.merger.SettingMerger`
        :param provider_name: name of the provider containing this value
        :param setting_name: name of the setting containing this value
        """
        value = merger.analyze_raw_value(self.value, provider_name, setting_name)
        value = os.path.normpath(value)
        if not value.endswith("/"):
            value += "/"
        if not os.path.isdir(value):
            self.add_warning(
                "'%s' is not a directory. Run the 'collectstatic' or 'migrate' command to fix this problem."
                % value,
                {"collectstatic", "migrate"},
            )
        return value

    def pre_collectstatic(self, merger, provider_name, setting_name, value):
        self.makedirs(merger, value)
        self.chmod(merger, value)

    def pre_migrate(self, merger, provider_name, setting_name, value):
        self.pre_collectstatic(merger, provider_name, setting_name, value)


class File(Path):
    """Represent a file name. Its parent directory is automatically created by the "migrate" and "collectstatic"
    command.

    """

    def get_value(self, merger, provider_name: str, setting_name: str):
        """Return the value

        :param merger: merger object, with all interpreted settings
        :type merger: :class:`df_config.config.merger.SettingMerger`
        :param provider_name: name of the provider containing this value
        :param setting_name: name of the setting containing this value
        """
        value = merger.analyze_raw_value(self.value, provider_name, setting_name)
        value = os.path.normpath(value)
        if not os.path.isfile(value):
            settings_check_results.append(
                Warning("'%s' does not exist." % value, obj="configuration")
            )
        return value

    def pre_collectstatic(self, merger, provider_name, setting_name, value):
        if value is None:
            return
        self.makedirs(merger, os.path.dirname(value))
        self.chmod(merger, value)

    def pre_migrate(self, merger, provider_name, setting_name, value):
        self.pre_collectstatic(merger, provider_name, setting_name, value)


class AutocreateFileContent(File):
    """Return the content of an existing file, or automatically write it and returns the content of the created file.
    Content must be a unicode string.
    The first argument of the provided `create_function` is the name of the file to create.

    The file is only written when the "migrate" Django command is called.
    The first arg of the provided `create_function` is a bool, in addition of your own *args and **kwargs:

        * `"collecstatic"` when Django is ready and your function is called for writing the file during
            the "migrate" command
        * `"migrate"` when Django is ready and your function is called for writing the file during the
            "collectstatic" command
        * `None` when Django is not ready and your function is called for loading settings

    """

    def __init__(
        self,
        value,
        create_function,
        mode=None,
        use_collectstatic: bool = False,
        use_migrate: bool = True,
        *args,
        **kwargs,
    ):
        """

        :param value: name of the file
        :param create_function: called when the file does not exist.
            Must return a text string.
        :param mode: mode of the create file (chmod value, like 0o755 for a file readable by everyone)
        :param use_collectstatic: the file is created by the 'collectstatic' command
        :param use_migrate: the file is created by the 'migrate' command
        :param args: extra args passed to the `create_function`
        :param kwargs:  extra keyword args passed to the `create_function`

        """
        super().__init__(value, mode=mode)
        self.create_function = create_function
        self.use_collectstatic = use_collectstatic
        self.use_migrate = use_migrate
        self.args = args
        self.kwargs = kwargs

    def create_file(self, merger, provider_name, setting_name, action="collectstatic"):
        filename = merger.analyze_raw_value(self.value, provider_name, setting_name)
        if filename is None or os.path.isfile(
            filename
        ):  # empty filename, or it already exists => nothing to do
            return
        result = self.create_function(action, *self.args, **self.kwargs)
        filename = os.path.abspath(filename)
        if result is not None:
            self.makedirs(merger, os.path.dirname(filename))
            merger.stdout.write("Writing new value to '%s'" % filename)
            try:
                result_text = self.serialize_value(result)
                with open(filename, "w") as fd:
                    fd.write(result_text)
            except Exception as e:
                merger.stderr.write(
                    "Unable to write content of '%s' (%s)" % (filename, e)
                )
            self.chmod(merger, filename)
        else:
            merger.stderr.write("Invalid empty content for '%s'" % filename)

    def pre_migrate(self, merger, provider_name, setting_name, value):
        if self.use_migrate:
            self.create_file(merger, provider_name, setting_name, action="migrate")

    def pre_collectstatic(self, merger, provider_name, setting_name, value):
        if self.use_collectstatic:
            self.create_file(
                merger, provider_name, setting_name, action="collectstatic"
            )

    def get_value(self, merger, provider_name: str, setting_name: str):
        """Return the value

        :param merger: merger object, with all interpreted settings
        :type merger: :class:`df_config.config.merger.SettingMerger`
        :param provider_name: name of the provider containing this value
        :param setting_name: name of the setting containing this value
        """
        filename = merger.analyze_raw_value(self.value, provider_name, setting_name)
        if os.path.isfile(filename):
            with open(filename) as fd:
                result_text = fd.read()
            result = self.unserialize_value(result_text)
        else:
            if self.use_collectstatic and self.use_migrate:
                bit = (
                    " Run the 'migrate' or 'collectstatic' command to fix this problem."
                )
                excluded = {"migrate", "collectstatic"}
            elif self.use_migrate:
                bit = " Run the 'migrate' command to fix this problem."
                excluded = {"migrate"}
            elif self.use_collectstatic:
                bit = " Run the 'collectstatic' command to fix this problem."
                excluded = {"collectstatic"}
            else:
                bit = ""
                excluded = set()

            self.add_warning(
                "'%s' does not exist.%s" % (filename, bit),
                excluded,
            )
            result = self.create_function(None, *self.args, **self.kwargs)
        return result

    # noinspection PyMethodMayBeStatic
    def serialize_value(self, value) -> str:
        """Serialize the result value to write it to the target file.

        :param value: the value returned by the `create_function`
        :return:
        """
        return value

    # noinspection PyMethodMayBeStatic
    def unserialize_value(self, value: str):
        """
        Format the text read in the target file.
        :param value: the content of the file
        :return:
        """
        return value


class AutocreateFile(AutocreateFileContent):
    """Represent a file name. If this file does not exist, an empty file is created by collectstatic/migrate commands"""

    def __init__(self, value, mode=None, *args, **kwargs):
        super().__init__(value, lambda x: "", mode=mode, *args, **kwargs)

    def get_value(self, merger, provider_name: str, setting_name: str):
        """Return the value

        :param merger: merger object, with all interpreted settings
        :type merger: :class:`df_config.config.merger.SettingMerger`
        :param provider_name: name of the provider containing this value
        :param setting_name: name of the setting containing this value
        """
        filename = merger.analyze_raw_value(self.value, provider_name, setting_name)
        if not os.path.isfile(filename):
            settings_check_results.append(
                Warning(
                    "'%s' does not exist. Run the 'migrate' command to fix this problem."
                    % filename,
                    obj="configuration",
                )
            )
        return filename


class SettingReference(DynamicSettting):
    """Reference any setting object by its name.
    Allow to reuse a list defined in another setting file.

    in `defaults.py`:

    >>> SETTING_1 = True
    >>> SETTING_2 = SettingReference('SETTING_1')

    In `local_settings.py`

    >>> SETTING_1 = False

    In your code:

    >>> from django.conf import settings

    Then `settings.SETTING_2` is equal to `False`
    """

    def __init__(self, value, func=None):
        super().__init__(value)
        self.func = func

    def get_value(self, merger, provider_name: str, setting_name: str):
        """Return the value

        :param merger: merger object, with all interpreted settings
        :type merger: :class:`df_config.config.merger.SettingMerger`
        :param provider_name: name of the provider containing this value
        :param setting_name: name of the setting containing this value
        """
        result = merger.get_setting_value(self.value)
        if self.func:
            result = self.func(result)
        return result


class CallableSetting(DynamicSettting):
    """
    Require a function(kwargs) as argument, this function will be called with all already computed settings in a dict.

    >>> SETTING_1 = True
    >>> def inverse_value(values):
    ...     # noinspection PyUnresolvedReferences
    ...     return not values['SETTING_1']
    >>> SETTING_2 = CallableSetting(inverse_value, 'SETTING_1')

    In `local_settings.py`

    >>> SETTING_1 = False

    In your code:

    >>> from django.conf import settings
    >>> if hasattr(settings, 'SETTING_2'):
    ...     assert(settings.SETTING_1 is False)  # default value is overriden by local_settings.py
    ...     assert(settings.SETTING_2 is True)  # SETTING_2 value is dynamically computed on startup

    Extra arguments must be strings, that are interpreted as required settings,
    that must be available before the call to your function. You can also set an attribute called
    `required_settings`.

    >>> def inverse_value(values):
    ...     # noinspection PyUnresolvedReferences
    ...     return not values['SETTING_1']
    >>> inverse_value.required_settings = ['SETTING_1']
    >>> SETTING_2 = CallableSetting(inverse_value)

    """

    def __init__(self, value, *required):
        if isinstance(value, str):
            value = import_string(value)
        super().__init__(value)
        if hasattr(value, "required_settings"):
            self.required = list(value.required_settings)
        else:
            self.required = required

    def get_value(self, merger, provider_name: str, setting_name: str):
        """Return the value

        :param merger: merger object, with all interpreted settings
        :type merger: :class:`df_config.config.merger.SettingMerger`
        :param provider_name: name of the provider containing this value
        :param setting_name: name of the setting containing this value
        """
        for required in self.required:
            merger.get_setting_value(required)
        value = self.value(merger.settings)
        return merger.analyze_raw_value(value, provider_name, setting_name)

    def __repr__(self):
        fn = repr(self.value)
        if hasattr(self.value, "__module__") and hasattr(self.value, "__name__"):
            fn = "%s.%s" % (self.value.__module__, self.value.__name__)
        return "CallableSetting(%r, %s)" % (
            fn,
            ", ".join(["%r" % x for x in self.required]),
        )


class ExpandIterable(SettingReference):
    """Allow to import an existing list inside a list setting.
    in `defaults.py`:

    >>> LIST_1 = [0, ]
    >>> LIST_2 = [1, ExpandIterable('LIST_1'), 2, ]
    >>> DICT_1 = {0: 0, }
    >>> DICT_2 = {1: 1, None: ExpandIterable('DICT_1'), 2: 2, }

    In case of dict, the key is ignored when the referenced dict is expanded.
    In `local_settings.py`

    >>> LIST_1 = [3, ]
    >>> DICT_1 = {3: 3, }

    In your code:

    >>> from django.conf import settings

    Then `settings.LIST_2` is equal to `[1, 3, 2]`.
    `settings.DICT_2` is equal to `{1: 1, 2: 2, 3: 3, }`.

    """

    pass



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
import copy
import os
import tempfile
from typing import Dict, Union
from unittest import TestCase

from df_config.checks import settings_check_results
from df_config.config.dynamic_settings import (
    AutocreateFile,
    AutocreateFileContent,
    CallableSetting,
    Directory,
    DynamicSettting,
    File,
    Path,
    RawValue,
    SettingReference,
)
from df_config.config.merger import SettingMerger
from df_config.config.values_providers import DictProvider


class TestDynamicSetting(TestCase):
    setting_name = "X"
    other_values: Dict[str, Union[str, DynamicSettting]] = {"OTHER": "42"}

    def check(
        self,
        dynamic_setting: DynamicSettting,
        expected_value,
        pre_collectstatic=False,
        pre_migrate=False,
        post_collectstatic=False,
        post_migrate=False,
        previous_settings=None,
    ):
        p_values = [x for x in settings_check_results]
        settings_check_results[:] = []
        values = previous_settings or copy.copy(self.other_values)
        values[self.setting_name] = dynamic_setting
        provider = DictProvider(values, name="d1")
        merger = SettingMerger(None, [provider])
        merger.load_raw_settings()
        actual_value = dynamic_setting.get_value(
            merger, provider.name, self.setting_name
        )
        self.assertEqual(expected_value, actual_value)
        if pre_collectstatic:
            dynamic_setting.pre_collectstatic(
                merger, provider.name, self.setting_name, actual_value
            )
        if pre_migrate:
            dynamic_setting.pre_migrate(
                merger, provider.name, self.setting_name, actual_value
            )
        if post_collectstatic:
            dynamic_setting.post_collectstatic(
                merger, provider.name, self.setting_name, actual_value
            )
        if post_migrate:
            dynamic_setting.post_migrate(
                merger, provider.name, self.setting_name, actual_value
            )
        n_values = [x for x in settings_check_results]
        settings_check_results[:] = p_values
        return n_values


class TestDynamicSettingClasses(TestDynamicSetting):
    def test_raw_value(self):
        self.check(RawValue("{X}"), "{X}")

    def test_path(self):
        self.check(Path("./test/../parent"), "parent")

    def test_directory(self):
        with tempfile.TemporaryDirectory() as dirname:
            path = dirname + "/test"
            r = self.check(
                Directory(path, mode=0o777),
                dirname + "/test/",
                pre_migrate=True,
                pre_collectstatic=True,
            )
            self.assertTrue(os.path.isdir(path))
            self.assertEqual(1, len(r))
            self.assertEqual(0o777, (os.stat(path).st_mode & 0o777))
            r = self.check(
                Directory(path, mode=0o700),
                dirname + "/test/",
                pre_migrate=True,
                pre_collectstatic=True,
            )
            self.assertTrue(os.path.isdir(path))
            self.assertEqual(0, len(r))
            self.assertEqual(0o700, (os.stat(path).st_mode & 0o777))

    def test_file(self):
        with tempfile.TemporaryDirectory() as dirname:
            path = dirname + "/test/file"
            r = self.check(
                File(path, mode=0o700), path, pre_migrate=True, pre_collectstatic=True
            )
            self.assertTrue(os.path.isdir(dirname + "/test"))
            self.assertFalse(os.path.isfile(path))
            self.assertEqual(1, len(r))
            with open(path, "w") as fd:
                fd.write("test")
            r = self.check(
                File(path, mode=0o700), path, pre_migrate=True, pre_collectstatic=True
            )
            self.assertTrue(os.path.isfile(path))
            self.assertEqual(0, len(r))
            self.assertEqual(0o700, (os.stat(path).st_mode & 0o777))

    def test_autocreatefilecontent(self):
        with tempfile.TemporaryDirectory() as dirname:
            path = dirname + "/test/file"
            r = self.check(
                AutocreateFileContent(path, lambda x: "test", mode=0o700),
                "test",
                pre_migrate=True,
                pre_collectstatic=True,
            )
            self.assertTrue(os.path.isfile(path))
            self.assertEqual(1, len(r))
            self.assertEqual(0o700, (os.stat(path).st_mode & 0o777))

    def test_autocreatefile(self):
        with tempfile.TemporaryDirectory() as dirname:
            path = dirname + "/test/file"
            r = self.check(
                AutocreateFile(path, mode=0o700),
                path,
                pre_migrate=True,
                pre_collectstatic=True,
            )
            self.assertTrue(os.path.isfile(path))
            self.assertEqual(1, len(r))
            self.assertEqual(0o700, (os.stat(path).st_mode & 0o777))

    def test_reference(self):
        self.check(SettingReference("OTHER"), "42")

    def test_callable_setting(self):
        def fn(values):
            return "[%(OTHER)s]" % values

        self.assertRaises(KeyError, lambda: self.check(CallableSetting(fn), "[42]"))
        self.check(CallableSetting(fn, "OTHER"), "[42]")
        fn.required_settings = ["OTHER"]
        self.check(CallableSetting(fn), "[42]")



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
import math
import os
import tempfile
from unittest import TestCase

from hypothesis import given
from hypothesis.strategies import floats, integers, text

from df_config.checks import settings_check_results
from df_config.config.fields import (
    BooleanConfigField,
    CharConfigField,
    ChoiceConfigFile,
    ConfigField,
    FloatConfigField,
    IntegerConfigField,
    ListConfigField,
    bool_setting,
    guess_relative_path,
    str_or_blank,
    str_or_none,
    strip_split,
)
from df_config.utils import ensure_dir


class TestFunctions(TestCase):
    def test_bool_setting(self):
        for k in {"1", "ok", "yes", "true", "on"}:
            self.assertTrue(bool_setting(k))
        for k in {"0", "ko", "no", "false", "of"}:
            self.assertFalse(bool_setting(k))

    @given(text())
    def test_bool_setting_multi(self, k):
        if k not in {"1", "ok", "yes", "true", "on"}:
            self.assertFalse(bool_setting(k))

    def test_str_or_none(self):
        self.assertIsNone(str_or_none(""))
        self.assertIsNone(str_or_none(None))

    @given(text())
    def test_str_or_none_multi(self, k):
        if k:
            self.assertEqual(k, str_or_none(k))

    def test_str_or_blank(self):
        self.assertEqual("", str_or_blank(""))
        self.assertEqual("", str_or_blank(None))

    @given(text())
    def test_str_or_blank_multi(self, k):
        self.assertEqual(k, str_or_blank(k))

    def test_guess_relative_path(self):
        cwd = os.getcwd()
        with tempfile.TemporaryDirectory() as dirname:
            os.chdir(dirname)
            dirname = os.getcwd()
            tested_path = os.path.join(dirname, "test1", "test2")
            r_1 = guess_relative_path(tested_path)
            ensure_dir(tested_path, parent=False)
            r_2 = guess_relative_path(tested_path)

        os.chdir(cwd)
        self.assertEqual("./test1/test2", r_1)
        self.assertEqual("./test1/test2", r_2)
        self.assertEqual("", guess_relative_path(None))
        self.assertEqual("", guess_relative_path(""))

    @given(text())
    def test_guess_relative_path_multi(self, value):
        guess_relative_path(value)

    def test_strip_split(self):
        r = strip_split("keyword1, keyword2 ,,keyword3")
        self.assertEqual(["keyword1", "keyword2", "keyword3"], r)
        self.assertEqual([], strip_split(",,,,"))
        self.assertEqual([], strip_split(None))
        self.assertEqual([], strip_split(""))


class TestFields(TestCase):
    def assertEqual(self, x, y, msg=None):
        if (
            isinstance(x, float)
            and isinstance(y, float)
            and math.isnan(x)
            and math.isnan(y)
        ):
            return
        super().assertEqual(x, y, msg=msg)

    def check(self, field: ConfigField, str_value, py_value, reverse: bool = True):
        self.assertEqual(py_value, field.from_str(str_value))
        if reverse:
            self.assertEqual(str_value, field.to_str(py_value))
            self.assertEqual(str_value, field.to_str(field.from_str(str_value)))
            self.assertEqual(py_value, field.from_str(field.to_str(py_value)))

    def test_char_config_field(self):
        self.check(CharConfigField("test.test", "TEST", allow_none=True), "", None)
        self.check(CharConfigField("test.test", "TEST", allow_none=False), "", "")

    @given(text())
    def test_char_config_field_multi(self, value):
        self.check(CharConfigField("test.test", "TEST", allow_none=False), value, value)

    def test_int_config_field(self):
        self.check(IntegerConfigField("test.test", "TEST", allow_none=False), "0", 0)
        self.check(IntegerConfigField("test.test", "TEST", allow_none=False), "1", 1)
        self.check(IntegerConfigField("test.test", "TEST", allow_none=True), "", None)
        self.check(
            IntegerConfigField("test.test", "TEST", allow_none=False),
            "",
            0,
            reverse=False,
        )

    @given(integers())
    def test__config_field_multi(self, value):
        self.check(
            IntegerConfigField("test.test", "TEST", allow_none=False), str(value), value
        )

    def test_float_config_field(self):
        self.check(FloatConfigField("test.test", "TEST", allow_none=False), "0.0", 0.0)
        self.check(FloatConfigField("test.test", "TEST", allow_none=False), "1.0", 1.0)
        self.check(FloatConfigField("test.test", "TEST", allow_none=True), "", None)
        self.check(
            FloatConfigField("test.test", "TEST", allow_none=False),
            "",
            0.0,
            reverse=False,
        )

    @given(floats())
    def test_float_config_field_multi(self, value):
        self.check(
            FloatConfigField("test.test", "TEST", allow_none=False), str(value), value
        )

    def test_list_config_field(self):
        self.check(ListConfigField("test.test", "TEST"), "0,1", ["0", "1"])
        self.check(ListConfigField("test.test", "TEST"), "", [])

    def test_bool_config_field(self):
        self.check(BooleanConfigField("test.test", "TEST"), "true", True)
        self.check(BooleanConfigField("test.test", "TEST"), "false", False)

    def test_choice_config_field(self):
        p_values = [x for x in settings_check_results]
        settings_check_results[:] = []
        self.check(
            ChoiceConfigFile("test.test", "TEST", {"1": "V1", "2": "V2"}), "1", "V1"
        )
        self.assertEqual(0, len(settings_check_results))
        self.check(
            ChoiceConfigFile("test.test", "TEST", {"1": "V1", "2": "V2"}),
            "3",
            None,
            reverse=False,
        )
        self.assertEqual(1, len(settings_check_results))

        settings_check_results[:] = p_values



# ##############################################################################
#  This file is part of Interdiode                                             #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <matthieu.gallet@19pouces.net>           #
#  All Rights Reserved                                                         #
#                                                                              #
# ##############################################################################
"""Root URLs provided by DjangoFloor
=================================

By default, register URLs for the admin site, `jsi18n`, static and media files, favicon and robots.txt.
If DjangoDebugToolbar is present, then its URL is also registered.

"""

from django.conf import settings
from django.conf.urls import include
from django.urls import path, re_path
from django.utils.module_loading import autodiscover_modules, import_string
from django.views.i18n import JavaScriptCatalog
from django.views.static import serve

from df_config.utils import get_view_from_string


def common_static_urls():
    values = (
        "robots.txt",
        "apple-touch-icon.png",
        "apple-touch-icon-precomposed.png",
        "favicon.ico",
    )
    return [
        path(
            filename,
            serve,
            kwargs={
                "document_root": settings.STATIC_ROOT,
                "path": "favicon/%s" % filename,
            },
        )
        for filename in values
    ]


catalog_view = JavaScriptCatalog.as_view(packages=settings.DF_JS_CATALOG_VIEWS)
urlpatterns = [
    path("jsi18n/", catalog_view, name="jsi18n"),
    path(
        "%s<path:path>" % settings.MEDIA_URL[1:],
        serve,
        name="serve_media",
        kwargs={"document_root": settings.MEDIA_ROOT},
    ),
    path(
        "%s<path:path>" % settings.STATIC_URL[1:],
        serve,
        name="serve_static",
        kwargs={"document_root": settings.STATIC_ROOT},
    ),
]

urlpatterns += common_static_urls()

if settings.DF_URL_CONF:
    try:
        extra_urls = import_string(settings.DF_URL_CONF)
        urlpatterns += list(extra_urls)
    except ModuleNotFoundError:
        pass

if settings.USE_ALL_AUTH:
    # noinspection PyUnresolvedReferences
    from allauth.account.views import login

    urlpatterns += [
        path("admin/login/", login),
        path("accounts/", include("allauth.urls")),
    ]
else:
    urlpatterns += [path("auth/", include("django.contrib.auth.urls"))]
if settings.USE_SITE:
    urlpatterns += [
        path("chaining/", include("smart_selects.urls")),
        path("df_site/", include("df_site.urls")),
    ]
if settings.DF_ADMIN_SITE:
    admin_site = import_string(settings.DF_ADMIN_SITE)
    autodiscover_modules("admin", register_to=admin_site)
    urlpatterns += [path("admin/", include(admin_site.urls[:2]))]
if settings.DEBUG and settings.USE_DEBUG_TOOLBAR:
    # noinspection PyPackageRequirements,PyUnresolvedReferences
    import debug_toolbar

    urlpatterns += [path("__debug__/", include(debug_toolbar.urls))]
if settings.DF_INDEX_VIEW:
    urlpatterns += [
        path("", get_view_from_string(settings.DF_INDEX_VIEW), name="index")
    ]
if settings.USE_WEBSOCKETS:
    from df_websockets.load import load_celery

    load_celery()

url_prefix = settings.URL_PREFIX[1:]

if url_prefix:
    urlpatterns = [re_path("^" + url_prefix, include(urlpatterns))]



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
import io
import os
import sys
from unittest import TestCase

from df_config.manage import (
    MODULE_VARIABLE_NAME,
    PYCHARM_VARIABLE_NAME,
    SETTINGS_VARIABLE_NAME,
    get_merger_from_env,
    manage,
    set_env,
)
from test_df_config.test_values_providers import EnvPatch


class TestSetEnv(TestCase):
    def test_set_env(self):
        delete = {PYCHARM_VARIABLE_NAME, SETTINGS_VARIABLE_NAME, MODULE_VARIABLE_NAME}

        with EnvPatch(delete=delete):
            module_name = set_env(module_name="Django-Floor")
            self.assertEqual(
                "df_config.config.base", os.environ[SETTINGS_VARIABLE_NAME]
            )
        self.assertEqual("django_floor", module_name)

        with EnvPatch(delete=delete, **{MODULE_VARIABLE_NAME: "demo"}):
            module_name = set_env(module_name="Django-Floor")
            self.assertEqual(
                "df_config.config.base", os.environ[SETTINGS_VARIABLE_NAME]
            )
        self.assertEqual("demo", module_name)

        with EnvPatch(delete=delete, **{SETTINGS_VARIABLE_NAME: "demo.settings"}):
            module_name = set_env(module_name="Django-Floor")
            self.assertEqual("demo.settings", os.environ[SETTINGS_VARIABLE_NAME])
        self.assertEqual("django_floor", module_name)

        with EnvPatch(delete=delete):
            module_name = set_env(argv=["demo-ctl"])
        self.assertEqual("demo", module_name)

        with EnvPatch(delete=delete):
            module_name = set_env(argv=["demo-ctl.py"])
        self.assertEqual("demo", module_name)

        with EnvPatch(delete=delete):
            module_name = set_env(argv=["demo-ctl.pyc"])
        self.assertEqual("demo", module_name)

    def test_manage(self):
        with EnvPatch(**{MODULE_VARIABLE_NAME: "df_config"}):
            sys.stdout = io.StringIO()
            sys.stderr = io.StringIO()
            manage(["df-config-ctl"])

    def test_get_merger_from_env(self):
        with EnvPatch(**{MODULE_VARIABLE_NAME: "df_config"}):
            merger = get_merger_from_env()
            self.assertEqual(8, len(merger.providers))
        with EnvPatch(**{MODULE_VARIABLE_NAME: "demo"}):
            merger = get_merger_from_env()
            self.assertEqual(8, len(merger.providers))



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
import argparse
import mimetypes
import os
import re
from email.utils import mktime_tz, parsedate_tz
from importlib import import_module
from typing import Iterable, Set, Tuple
from urllib.parse import quote

import pkg_resources
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import (
    HttpRequest,
    HttpResponse,
    HttpResponseNotModified,
    StreamingHttpResponse,
)
from django.utils.http import http_date
from django.utils.module_loading import import_string


class RemovedInDjangoFloor200Warning(DeprecationWarning):
    """Used for displaying functions or modules that will be removed in a near future."""

    pass


def ensure_dir(path, parent=True):
    """Ensure that the given directory exists

    :param path: the path to check
    :param parent: only ensure the existence of the parent directory

    """
    dirname = os.path.dirname(path) if parent else path
    os.makedirs(dirname, exist_ok=True)
    return path


def is_package_present(package_name):
    """Return True is the `package_name` package is present in your current Python environment."""
    try:
        import_module(package_name)
        return True
    except ImportError:
        return False


def remove_arguments_from_help(parser: argparse.ArgumentParser, arguments: Set):
    # noinspection PyProtectedMember
    for action in parser._actions:
        if arguments & set(action.option_strings):
            action.help = argparse.SUPPRESS


def guess_version(defined_settings):
    """Guesss the project version.
    Expect an installed version (findable with pkg_resources) or __version__ in `your_project/__init__.py`.
    If not found

    :param defined_settings: all already defined settings (dict)
    :type defined_settings: :class:`dict`
    :return: should be something like `"1.2.3"`
    :rtype: :class:`str`
    """
    try:
        project_distribution = pkg_resources.get_distribution(
            defined_settings["DF_MODULE_NAME"]
        )
        return project_distribution.version
    except pkg_resources.DistributionNotFound:
        pass
    try:
        return import_string("%s.__version__" % defined_settings["DF_MODULE_NAME"])
    except ImportError:
        return "1.0.0"


def get_view_from_string(view_as_str):
    try:
        view = import_string(view_as_str)
    except ImportError:
        raise ImproperlyConfigured("Unable to import %s" % view_as_str)
    if hasattr(view, "as_view") and callable(view.as_view):
        return view.as_view()
    elif callable(view):
        return view
    raise ImproperlyConfigured(
        '%s is not callabled and does not have an "as_view" attribute'
    )


class ChunkReader:
    """read a file object in chunks of the given size.

    Return an iterator of data

    :param fileobj:
    :param chunk_size: max size of each chunk
    :type chunk_size: `int`
    """

    def __init__(self, fileobj, chunk_size=32768):
        self.fileobj = fileobj
        self.chunk_size = chunk_size

    def __iter__(self):
        for data in iter(lambda: self.fileobj.read(self.chunk_size), b""):
            yield data

    def close(self):
        self.fileobj.close()


class RangedChunkReader(ChunkReader):
    def __init__(self, fd, ranges: Iterable[Tuple[int, int]], chunk_size=32768):
        super().__init__(fd, chunk_size=chunk_size)
        self.ranges = ranges

    def __iter__(self):
        for start, end in self.ranges:
            self.fileobj.seek(start)
            while start <= end:
                size = min(self.chunk_size, end - start + 1)
                if size <= 0:
                    break
                yield self.fileobj.read(size)
                start += size


mimetypes.init()


_range_re = re.compile(r"^bytes=(\d*-\d*(?:,\s*\d*-\d*)*)$")
_if_modified_since_re = re.compile(
    r"^([^;]+)(; length=([0-9]+))?$", flags=re.IGNORECASE
)


def was_modified_since(header=None, mtime=0.0, size=None):
    """
    Was something modified since the user last downloaded it?
    header
      This is the value of the If-Modified-Since header.  If this is None,
      I'll just return True.
    mtime
      This is the modification time of the item we're talking about.
    size
      This is the size of the item we're talking about.
    """
    if header is None:
        return True
    matcher = _if_modified_since_re.match(header)
    if not matcher:
        return True
    try:
        header_date = parsedate_tz(matcher.group(1))
        if header_date is None:
            return True
        header_mtime = mktime_tz(header_date)
        header_len = matcher.group(3)
        if size is not None and header_len and int(header_len) != size:
            return True
        if mtime > header_mtime:
            return True
    except (AttributeError, ValueError, OverflowError):
        return True
    return False


def send_file(
    request: HttpRequest,
    filepath: str,
    mimetype=None,
    force_download=False,
    attachment_filename=None,
    chunk_size=32768,
    etag=None,
):
    """Send a local file. This is not a Django view, but a function that is called at the end of a view.

    If `settings.USE_X_SEND_FILE` (mod_xsendfile is a mod of Apache), then return an empty HttpResponse with the
    correct header. The file is directly handled by Apache instead of Python.
    If `settings.X_ACCEL_REDIRECT_ARCHIVE` is defined (as a list of tuple (directory, alias_url)) and filepath is
    in one of the directories, return an empty HttpResponse with the correct header.
    This is only available with Nginx.

    Otherwise, return a StreamingHttpResponse to avoid loading the whole file in memory.

    :param request: the original request (used to detect the "range" header)
    :param filepath: absolute path of the file to send to the client.
    :param mimetype: MIME type of the file (returned in the response header)
    :param force_download: always force the client to download the file.
    :param attachment_filename: filename used in the "Content-Disposition" header (when used)
    :param chunk_size: size of chunks for large files. Useful at least for unittests
    :param etag: ETag header to add to the response
    :rtype: :class:`django.http.response.StreamingHttpResponse` or :class:`django.http.response.HttpResponse`
    """
    if mimetype is None:
        (mimetype, encoding) = mimetypes.guess_type(filepath)
        if mimetype is None:
            mimetype = "text/plain"
    if isinstance(mimetype, bytes):
        # noinspection PyTypeChecker
        mimetype = mimetype.decode("utf-8")

    filepath = os.path.abspath(filepath)

    attachment_filename = attachment_filename or os.path.basename(filepath)
    if request.method == "HEAD":
        r = HttpResponse(content=b"", content_type=mimetype, status=200)
        if etag:
            r["ETag"] = str(etag)
        return r

    range_matcher = _range_re.match(request.META.get("HTTP_RANGE", ""))
    ranges = []
    if not os.path.isfile(filepath):
        return HttpResponse(status=404)
    stats = os.stat(filepath)
    filesize = stats.st_size
    if not was_modified_since(
        request.META.get("HTTP_IF_MODIFIED_SINCE"), mtime=stats.st_mtime, size=filesize
    ):
        return HttpResponseNotModified()
    if range_matcher:
        content_size = 0
        ranges_str = [x.strip() for x in range_matcher.group(1).split(",")]
        for range_str in ranges_str:
            start_str, sep, end_str = range_str.partition("-")
            end = int(end_str) if end_str else filesize - 1
            start = int(start_str) if start_str else filesize - end
            content_size += end - start + 1
            if end + 1 > filesize:
                response = HttpResponse(content_type=mimetype, status=416)
                response["Content-Range"] = "bytes */%s" % filesize

                return response
            ranges.append((start, end))
    else:
        content_size = filesize

    response = None
    if settings.USE_X_SEND_FILE and not ranges:
        response = HttpResponse(content_type=mimetype)
        response["X-SENDFILE"] = filepath
        response["Content-Length"] = str(content_size)
    elif settings.X_ACCEL_REDIRECT and not ranges:
        for dirpath, alias_url in settings.X_ACCEL_REDIRECT:
            dirpath = os.path.abspath(dirpath)
            if filepath.startswith(dirpath):
                response = HttpResponse(content_type=mimetype)
                response["X-Accel-Redirect"] = os.path.join(
                    alias_url, os.path.relpath(filepath, dirpath)
                )
                response["Content-Length"] = str(content_size)
                break
    if response is None:
        fileobj = open(filepath, "rb")
        status = 200
        if ranges:
            file_content = RangedChunkReader(fileobj, ranges, chunk_size=chunk_size)
            if len(ranges) == 1:
                status = 206
        else:
            file_content = ChunkReader(fileobj, chunk_size=chunk_size)
        response = StreamingHttpResponse(
            file_content, content_type=mimetype, status=status
        )
        if len(ranges) == 1:
            response["Content-Range"] = "bytes %d-%d/%d" % (
                ranges[0][0],
                ranges[0][1],
                filesize,
            )
        response["Content-Length"] = content_size
    response["Last-Modified"] = http_date(stats.st_mtime)
    encoded_filename = quote(attachment_filename, encoding="utf-8")
    header = "attachment" if force_download else "inline"
    if etag:
        response["ETag"] = str(etag)
    if encoded_filename == attachment_filename:
        response["Content-Disposition"] = '{1}; filename="{0}"'.format(
            encoded_filename, header
        )
    else:
        response["Content-Disposition"] = "{1};filename*=UTF-8''\"{0}\"".format(
            encoded_filename, header
        )
    return response



import re
import urllib.parse
from typing import List, Optional

from pkg_resources import DistributionNotFound, get_distribution

from df_config.config.dynamic_settings import DynamicSettting


class Attribute(DynamicSettting):
    def __init__(self, parsed_url, attribute: callable, default=None):
        super().__init__(attribute)
        self.url_setting: URLSetting = parsed_url
        self.default = default

    def get_value(self, merger, provider_name: str, setting_name: str):
        self.url_setting.load(merger)
        if self.url_setting.parsed_url:
            value = self.value()
        else:
            value = self.default
        return merger.analyze_raw_value(value, provider_name, setting_name)

    def __repr__(self):
        method = self.value.__name__[:-1]
        default = self.default
        setting = (
            f"{self.url_setting.__class__.__name__}({self.url_setting.setting_name!r})"
        )
        return f"{setting}.{method}(default={default!r})"


class URLSetting:
    ENGINES = {}
    REQUIREMENTS = {}
    SCHEME_ALIASES = {}
    SCHEMES = {
        "amqp": (5672, False, False),
        "file": (None, False, False),
        "http": (80, False, False),
        "https": (443, True, False),
        "memcache": (11211, False, False),
        "mariadb": (3306, False, False),
        "mysql": (3306, False, False),
        "oracle": (1521, False, False),
        "postgres": (5432, False, False),
        "redis": (6379, False, False),
        "rediss": (6379, True, False),
        "smtp": (25, False, False),
        "smtp+tls": (487, False, True),
        "smtps": (465, True, False),
        "sqlite": (None, False, False),
        "sqlite3": (None, False, False),
    }

    def __init__(
        self,
        setting_name: str = None,
        required: List[str] = None,
        url: Optional[str] = None,
    ):
        if url is None:
            self.parsed_url = None
            self._url_str = None
            self._loaded = False
        else:
            self._url_str = url
            self.parsed_url = urllib.parse.urlparse(self._url_str)
            self._loaded = True
        self.setting_name = setting_name
        self.required = required or []

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.setting_name}')"

    def load(self, merger):
        if self._loaded or not self.setting_name:
            return
        for required in self.required:
            merger.get_setting_value(required)
        value = merger.get_setting_value(self.setting_name)
        if value is self or not value:
            self._url_str = None
            self.parsed_url = None
        else:
            self._url_str = value
            self.parsed_url = urllib.parse.urlparse(self._url_str)
        self._loaded = True

    def hostname(self, default="localhost"):
        return Attribute(self, self.hostname_, default=default)

    def hostname_(self):
        return self.parsed_url.hostname if self.parsed_url else None

    def netloc(self, default="localhost"):
        return Attribute(self, self.netloc_, default=default)

    def netloc_(self):
        return self.parsed_url.netloc if self.parsed_url else None

    def params(self, default=""):
        return Attribute(self, self.params_, default=default)

    def params_(self):
        return self.parsed_url.params if self.parsed_url else None

    def password(self, default=None):
        return Attribute(self, self.password_, default=default)

    def password_(self):
        return self.parsed_url.password if self.parsed_url else None

    def path(self, default=""):
        return Attribute(self, self.path_, default=default)

    def path_(self):
        return self.parsed_url.path if self.parsed_url else None

    def port(self, default=None):
        return Attribute(self, self.port_, default=default)

    def port_(self):
        return self.parsed_url.port if self.parsed_url else None

    def query(self, default=""):
        return Attribute(self, self.query_, default=default)

    def query_(self):
        return self.parsed_url.query if self.parsed_url else None

    def scheme(self, default=None):
        return Attribute(self, self.scheme_, default=default)

    def scheme_(self):
        if self.parsed_url is None:
            return None
        s = self.parsed_url.scheme.lower()
        return self.SCHEME_ALIASES.get(s, s)

    def username(self, default=None):
        return Attribute(self, self.username_, default=default)

    def username_(self):
        return self.parsed_url.username if self.parsed_url else None

    def database(self, default=None):
        return Attribute(self, self.database_, default=default)

    def database_(self):
        if self.parsed_url is None or not self.parsed_url.path:
            return None
        matcher = re.match(r"/([^/]+)", self.parsed_url.path)
        if not matcher:
            return None
        return matcher.group(1)

    def port_int(self, default=None):
        """Return an integer value for the port (the default one if not specified)."""
        return Attribute(self, self.port_int_, default=default)

    def port_int_(self) -> Optional[int]:
        if not self.parsed_url:
            return None
        if self.parsed_url.port:
            return int(self.parsed_url.port)
        s = self.scheme_()
        return self.SCHEMES[s][0]

    def use_tls(self, default=False):
        return Attribute(self, self.use_tls_, default=default)

    def use_tls_(self):
        if not self.parsed_url:
            return False
        s = self.scheme_()
        return self.SCHEMES[s][2]

    def use_ssl(self, default=False):
        return Attribute(self, self.use_ssl_, default=default)

    def use_ssl_(self):
        if not self.parsed_url:
            return False
        s = self.scheme_()
        return self.SCHEMES[s][1]

    def engine(self, default=None):
        return Attribute(self, self.engine_, default=default)

    def engine_(self):
        if not self.parsed_url:
            return None
        return self.normalize_engine(self.scheme_())

    @classmethod
    def normalize_engine(cls, scheme):
        engine = cls.ENGINES.get(scheme, scheme)
        requirements = cls.REQUIREMENTS.get(engine, [])
        found = False
        for req in requirements:
            try:
                get_distribution(req)
                found = True
            except DistributionNotFound:
                pass
        if not found and requirements:
            from df_config.checks import missing_package, settings_check_results

            settings_check_results.append(
                missing_package("/".join(requirements), f" to use {engine}.")
            )
        return engine


class DatabaseURL(URLSetting):
    ENGINES = {
        "mysql": "django.db.backends.mysql",
        "mariadb": "django.db.backends.mysql",
        "oracle": "django.db.backends.oracle",
        "postgres": "django.db.backends.postgresql",
        "sqlite3": "django.db.backends.sqlite3",
    }
    SCHEME_ALIASES = {
        "psql": "postgres",
        "postgresql": "postgres",
        "sqlite": "sqlite3",
    }
    REQUIREMENTS = {
        "django.db.backends.postgresql": ["psycopg2-binary", "psycopg2"],
        "django.db.backends.oracle": ["cx_Oracle"],
        "django.db.backends.mysql": ["mysqlclient", "pymysql"],
    }


class RedisURL(URLSetting):
    def database_(self):
        if self.parsed_url is None:
            return None
        elif not self.parsed_url.path:
            return 1
        matcher = re.match(r"/(0|[1-9]\d*)$", self.parsed_url.path)
        if not matcher:
            return 0
        return int(matcher.group(1))



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
from importlib import import_module
from typing import Any, List, Optional, Tuple

from df_config.config.fields import ConfigField


def import_attribute(value: str, default=None) -> Tuple[Optional[Any], bool]:
    if value is None:
        return default, False
    module_name, sep, attribute_name = value.partition(":")
    try:
        module = import_module(module_name, package=None)
        return getattr(module, attribute_name), True
    except AttributeError:
        return default, False
    except ImportError:
        return default, False


class ConfigFieldsProvider:
    """Provides a list of :class:`df_config.config.fields.ConfigField`.
    Used for retrieving settings from a config file.
    """

    name = None

    def get_config_fields(self) -> List[ConfigField]:
        """Return a list of config fields to search in values providers"""
        raise NotImplementedError

    def is_valid(self) -> bool:
        """return True if the provider is working"""
        raise NotImplementedError


class PythonConfigFieldsProvider(ConfigFieldsProvider):
    """Provide a list of :class:`df_config.config.fields.ConfigField` from an attribute in a Python module."""

    name = "Python attribute"

    def __init__(self, value=None, fallback=None):
        self.attribute_name = value or "df_config.iniconf:EMPTY_INI_MAPPING"
        self.mapping, self.valid = import_attribute(self.attribute_name, [])
        if not self.valid and fallback:
            self.attribute_name = fallback
            self.mapping, self.valid = import_attribute(self.attribute_name, [])

    def is_valid(self) -> bool:
        return self.valid

    def get_config_fields(self):
        """Return the list that is defined in the module by the attribute name"""
        return self.mapping

    def __str__(self):
        return self.attribute_name



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
from django.apps import AppConfig
from django.db.models.signals import post_migrate, pre_migrate

# noinspection PyUnresolvedReferences
import df_config.checks
from df_config.config.base import merger


class MigrationCounter:
    def __init__(self):
        self.last_app_config = None

    # noinspection PyUnusedLocal
    def pre_migrate(self, *args, app_config: AppConfig = None, **kwargs):
        if self.last_app_config is None:  # this is the pre_migrate of first app
            merger.call_method_on_config_values("pre_migrate")
        self.last_app_config = app_config.name

    # noinspection PyUnusedLocal
    def post_migrate(self, *args, app_config: AppConfig = None, **kwargs):
        if (
            app_config.name == self.last_app_config
        ):  # this is the post_migrate of the last app
            merger.call_method_on_config_values("post_migrate")


migration_counter = MigrationCounter()

pre_migrate.connect(migration_counter.pre_migrate)
post_migrate.connect(migration_counter.post_migrate)



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
import ipaddress
import logging
import os
import re
import sys
from typing import List

from df_config.config.fields_providers import PythonConfigFieldsProvider
from df_config.config.merger import SettingMerger
from df_config.config.values_providers import (
    DictProvider,
    EnvironmentConfigProvider,
    IniConfigProvider,
    PythonFileProvider,
    PythonModuleProvider,
)

PYCHARM_VARIABLE_NAME = "PYCHARM_DJANGO_MANAGE_MODULE"
SETTINGS_VARIABLE_NAME = "DJANGO_SETTINGS_MODULE"
MODULE_VARIABLE_NAME = "DF_CONF_NAME"


def set_env(module_name: str = None, argv: List[str] = None):
    """Set the environment variable `DF_CONF_NAME` with the main Python module name
    The value looks like "project_name".
    If `module_name` is not given, tries to infer it from the running script name

    """
    if MODULE_VARIABLE_NAME not in os.environ:
        script_re = re.compile(r"^([\w_\-.]+)-\w+(?:\.py|\.pyc|)$")
        if not module_name:
            if PYCHARM_VARIABLE_NAME in os.environ:
                pycharm_matcher = script_re.match(os.environ[PYCHARM_VARIABLE_NAME])
                if pycharm_matcher:
                    module_name = pycharm_matcher.group(1)
        if not module_name:
            argv = argv or sys.argv
            if argv and argv[0]:
                script_matcher = script_re.match(os.path.basename(argv[0]))
                if script_matcher:
                    module_name = script_matcher.group(1)
        if not module_name:
            module_name = "df_config"
        os.environ[MODULE_VARIABLE_NAME] = module_name.replace("-", "_").lower()
    os.environ.setdefault(SETTINGS_VARIABLE_NAME, "df_config.config.base")
    return os.environ[MODULE_VARIABLE_NAME]


def get_merger_from_env(merger_class=SettingMerger) -> SettingMerger:
    """Return a settingmerger to determien all available settings, should be used after set_env().
    Settings are found in this order:

    * df_config.config.defaults
    * {project_name}.defaults (overrides df_config.config.defaults)
    * {root}/etc/{project_name}/settings.ini (overrides {project_name}.settings)
    * {root}/etc/{project_name}/settings.py (overrides {root}/etc/{project_name}/settings.ini)
    * ./local_settings.ini (overrides {root}/etc/{project_name}/settings.py)
    * ./local_settings.py (overrides ./local_settings.ini)
    * environment variables (overrides ./local_settings.py)
    """
    # required if set_env is not called
    module_name = set_env()
    prefix = os.path.abspath(sys.prefix)
    if prefix == "/usr":
        prefix = ""
    ini_mapping = f"{module_name}.iniconf:INI_MAPPING"
    config_providers = [
        DictProvider({"DF_MODULE_NAME": module_name}, name="default values"),
        PythonModuleProvider("df_config.config.defaults"),
        PythonModuleProvider("%s.defaults" % module_name),
        IniConfigProvider(
            "%s/etc/%s/settings.ini"
            % (
                prefix,
                module_name,
            )
        ),
        PythonFileProvider(
            "%s/etc/%s/settings.py"
            % (
                prefix,
                module_name,
            )
        ),
        EnvironmentConfigProvider("%s_" % module_name.upper()),
        IniConfigProvider(os.path.abspath("local_settings.ini")),
        PythonFileProvider(os.path.abspath("local_settings.py")),
    ]
    fields_provider = PythonConfigFieldsProvider(
        ini_mapping, fallback="df_config.iniconf:DEFAULT_INI_MAPPING"
    )
    return merger_class(fields_provider, config_providers)


def manage(argv=None):
    set_env()
    import django

    django.setup()
    from django.core.management import execute_from_command_line

    patch_commands()
    argv = argv or sys.argv
    logger = logging.getLogger("django.server")
    logger.info("command='%s'", " ".join(argv))
    execute_from_command_line(argv=argv)


def patch_commands():
    """patch the runserver command to use the configured LISTEN_ADDRESS"""
    from django.conf import settings
    from django.core.management.commands.runserver import Command

    if not hasattr(settings, "LISTEN_ADDRESS"):
        return
    add, sep, port = settings.LISTEN_ADDRESS.rpartition(":")
    if sep == ":":
        try:
            Command.default_port = str(int(port))
            add = ipaddress.ip_address(add)
            if add.version == 4:
                Command.default_addr = str(add)
            else:
                Command.default_addr_ipv6 = str(add)
        except ValueError:
            pass



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################


# noinspection PyClassHasNoInit,PyAbstractClass
import os
import subprocess
from pathlib import Path

from django.conf import settings

if settings.USE_PIPELINE:

    # noinspection PyPackageRequirements,PyUnresolvedReferences
    # noinspection PyPackageRequirements,PyUnresolvedReferences
    from pipeline.compilers import CompilerBase
    from pipeline.compressors import CompressorBase, SubProcessCompressor

    # noinspection PyPackageRequirements,PyUnresolvedReferences
    from pipeline.storage import PipelineManifestStorage, PipelineMixin
else:
    CompressorBase = object
    CompilerBase = object
    SubProcessCompressor = object
    PipelineManifestStorage = None
    PipelineMixin = None


if settings.USE_WHITENOISE:
    # noinspection PyPackageRequirements,PyUnresolvedReferences
    from whitenoise.storage import CompressedManifestStaticFilesStorage
else:
    CompressedManifestStaticFilesStorage = None


class RcssCompressor(CompressorBase):
    """
    CSS compressor based on the Python library slimit
    (https://github.com/ndparker/rcssmin).
    """

    def filter_css(self, css):
        raise NotImplementedError

    def filter_js(self, js):
        raise NotImplementedError

    # noinspection PyMethodMayBeStatic
    def compress_css(self, css):
        # noinspection PyUnresolvedReferences,PyPackageRequirements
        from rcssmin import cssmin

        return cssmin(css)


class CssNanoCompressor(SubProcessCompressor):
    def compress_css(self, css):
        command = [settings.CSSNANO_BINARY] + settings.CSSNANO_ARGUMENTS
        return self.execute_command(command, css)

    def filter_css(self, css):
        raise NotImplementedError

    def filter_js(self, js):
        raise NotImplementedError


class TerserCompressor(SubProcessCompressor):
    def compress_js(self, js):
        command = [settings.TERSER_BINARY, settings.TERSER_ARGUMENTS, "--"]
        if self.verbose:
            command += ["--verbose"]
        return self.execute_command(command, js)

    def filter_css(self, css):
        raise NotImplementedError

    def filter_js(self, js):
        raise NotImplementedError


# noinspection PyClassHasNoInit
class PyScssCompiler(CompilerBase):
    """SASS (.scss) compiler based on the Python library pyScss.
    (http://pyscss.readthedocs.io/en/latest/ ).
    However, this compiler is limited to SASS 3.2 and cannot compile modern projets like Bootstrap 4.
    Please use :class:`pipeline.compilers.sass.SASSCompiler` if you use modern SCSS files.

    """

    output_extension = "css"

    # noinspection PyMethodMayBeStatic
    def match_file(self, filename):
        return filename.endswith(".scss") or filename.endswith(".sass")

    # noinspection PyUnusedLocal
    def compile_file(self, infile, outfile, outdated=False, force=False):
        # noinspection PyUnresolvedReferences,PyUnresolvedReferences,PyPackageRequirements
        from scss import Compiler

        root = Path(os.path.abspath(settings.STATIC_ROOT))
        compiler = Compiler(root=root, search_path=("./",))
        css_content = compiler.compile(infile)
        with open(outfile, "w") as fd:
            fd.write(css_content)
        # noinspection PyUnresolvedReferences
        if self.verbose:
            print(css_content)


# noinspection PyClassHasNoInit
class TypescriptCompiler(CompilerBase):
    """TypeScript (.ts) compiler using "tsc".
    (https://www.typescriptlang.org ).

    """

    output_extension = "js"

    # noinspection PyMethodMayBeStatic
    def match_file(self, filename):
        return filename.endswith(".ts")

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def compile_file(self, infile, outfile, outdated=False, force=False):
        # noinspection PyPackageRequirements,PyUnresolvedReferences
        from pipeline.exceptions import CompilerError

        command = (
            [settings.TYPESCRIPT_BINARY]
            + settings.TYPESCRIPT_ARGUMENTS
            + ["-out", outfile, infile]
        )
        try:
            p = subprocess.Popen(
                command,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            stdout, __ = p.communicate(b"")
            if p.returncode != 0:
                raise CompilerError(
                    "Unable to execute TypeScript",
                    command=command,
                    error_output=stdout.decode(),
                )
        except Exception as e:
            raise CompilerError(e, command=command, error_output=str(e))


if PipelineManifestStorage:

    class NicerPipelineCachedStorage(PipelineManifestStorage):
        """display a better exception"""

        def hashed_name(self, name, content=None, filename=None):
            try:
                return super().hashed_name(name, content=content)
            except ValueError as e:
                raise ValueError(
                    "%s. Did you run the command 'collectstatic'?" % e.args[0]
                )

else:
    NicerPipelineCachedStorage = None

if PipelineMixin and CompressedManifestStaticFilesStorage:

    class PipelineCompressedManifestStaticFilesStorage(
        PipelineMixin, CompressedManifestStaticFilesStorage
    ):
        """mix django-pipeline and whitenoise"""

        pass

else:
    PipelineCompressedManifestStaticFilesStorage = None



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
from typing import List

from df_config.config.fields import (
    BooleanConfigField,
    CharConfigField,
    ChoiceConfigFile,
    ConfigField,
    FloatConfigField,
    IntegerConfigField,
    ListConfigField,
    bool_setting,
)


def x_accel_converter(value):
    if bool_setting(value):
        return [("{MEDIA_ROOT}/", "{MEDIA_URL}")]
    return []


ALLAUTH_MAPPING = []
BASE_MAPPING = [
    CharConfigField(
        "global.admin_email",
        "ADMIN_EMAIL",
        help_str="e-mail address for receiving logged errors",
        env_name="ADMIN_EMAIL",
    ),
    CharConfigField(
        None,
        "SECRET_KEY",
        help_str="Secret key to provide cryptographic signing, and should be set to a unique, unpredictable value.",
        env_name="SECRET_KEY",
    ),
    CharConfigField(
        "global.data",
        "LOCAL_PATH",
        help_str="where all data will be stored (static/uploaded/temporary files, …). "
        "If you change it, you must run the collectstatic and migrate commands again.\n",
    ),
    CharConfigField(
        "global.upload_directory",
        "MEDIA_ROOT",
        help_str="where uploaded media files will be stored.",
        env_name="UPLOAD_DIRECTORY",
    ),
    CharConfigField(
        "global.data",
        "FILE_UPLOAD_TEMP_DIR",
        help_str="where temporary data will be stored.",
        env_name="TEMP_DIRECTORY",
    ),
    CharConfigField(
        "global.language_code",
        "LANGUAGE_CODE",
        help_str="default to fr_FR",
        env_name="LANGUAGE_CODE",
    ),
    CharConfigField(
        "global.listen_address",
        "LISTEN_ADDRESS",
        help_str="address used by your web server (like 127.0.0.1:8000).",
    ),
    IntegerConfigField(
        None,
        "LISTEN_PORT",
        help_str="port to listen to (force to listen on 0.0.0.0:$PORT).",
        env_name="PORT",
    ),
    CharConfigField(
        "global.server_url",
        "SERVER_BASE_URL",
        help_str="Public URL of your website. \n"
        'Default to "http://{listen_address}/" but should '
        "be different if you use a reverse proxy like "
        "Apache or Nginx. Example: http://www.example.org/.",
        env_name="SERVER_URL",
    ),
    CharConfigField(
        None,
        "HEROKU_APP_NAME",
        help_str="Heroku's application name, when deploying your on Heroku.",
        env_name="HEROKU_APP_NAME",
    ),
    CharConfigField(
        "global.time_zone",
        "TIME_ZONE",
        help_str="default to Europe/Paris",
        env_name="TIME_ZONE",
    ),
    CharConfigField(
        None,
        "EMAIL_HOST_URL",
        help_str="SMTP server for sending admin emails. \n"
        "smtp+tls://account@example.com:password@smtp.example.com:587/",
        env_name="EMAIL_HOST_URL",
    ),
    CharConfigField("email.host", "EMAIL_HOST", help_str="SMTP server"),
    CharConfigField("email.password", "EMAIL_HOST_PASSWORD", help_str="SMTP password"),
    IntegerConfigField(
        "email.port", "EMAIL_PORT", help_str="SMTP port (often 25, 465 or 587)"
    ),
    CharConfigField("email.user", "EMAIL_HOST_USER", help_str="SMTP user"),
    CharConfigField("email.from", "EMAIL_FROM", help_str="Displayed sender email"),
    BooleanConfigField(
        "email.use_tls",
        "EMAIL_USE_TLS",
        help_str='"true" if your SMTP uses STARTTLS ' "(often on port 587)",
    ),
    BooleanConfigField(
        "email.use_ssl",
        "EMAIL_USE_SSL",
        help_str='"true" if your SMTP uses SSL (often on port 465)',
    ),
]
CELERY_MAPPING = [
    IntegerConfigField(
        "celery.db",
        "CELERY_DB",
        help_str="Database number of the Redis Celery DB\n"
        "Celery is used for processing background tasks and websockets.",
    ),
    CharConfigField("celery.host", "CELERY_HOST", help_str="Redis Celery DB host"),
    CharConfigField(
        "celery.password",
        "CELERY_PASSWORD",
        help_str="Redis Celery DB password (if required)",
    ),
    IntegerConfigField("celery.port", "CELERY_PORT", help_str="Redis Celery DB port"),
    IntegerConfigField(
        "celery.processes", "CELERY_PROCESSES", help_str="number of Celery processes"
    ),
]
DATABASE_MAPPING = [
    CharConfigField(
        None, "DATABASE_URL", help_str="URL of the database", env_name="DATABASE_URL"
    ),
    CharConfigField(
        "database.db",
        "DATABASE_NAME",
        help_str="Database name (or path of the sqlite3 database)",
        env_name=None,
    ),
    CharConfigField(
        "database.engine",
        "DATABASE_ENGINE",
        help_str='Database engine ("mysql", "postgresql", "sqlite3", "oracle", or the dotted name of '
        "the Django backend)",
        env_name=None,
    ),
    CharConfigField(
        "database.host", "DATABASE_HOST", help_str="Database host", env_name=None
    ),
    CharConfigField(
        "database.password",
        "DATABASE_PASSWORD",
        help_str="Database password",
        env_name=None,
    ),
    IntegerConfigField(
        "database.port", "DATABASE_PORT", help_str="Database port", env_name=None
    ),
    CharConfigField(
        "database.user", "DATABASE_USER", help_str="Database user", env_name=None
    ),
]
DJANGO_AUTH_MAPPING = [
    BooleanConfigField(
        "auth.local_users",
        "DF_ALLOW_LOCAL_USERS",
        help_str='Set to "false" to deactivate local database of users.',
    ),
    BooleanConfigField(
        "auth.create_users",
        "DF_ALLOW_USER_CREATION",
        help_str='Set to "false" if users cannot create their account themselvers, or '
        "only if existing users can by authenticated by the reverse-proxy.",
    ),
    IntegerConfigField(
        "auth.session_duration",
        "SESSION_COOKIE_AGE",
        help_str="Duration of the connection sessions "
        "(in seconds, default to 1,209,600 s / 14 days)",
    ),
    BooleanConfigField(
        "auth.allow_basic_auth",
        "USE_HTTP_BASIC_AUTH",
        help_str='Set to "true" if you want to allow HTTP basic auth, using the Django database.',
    ),
]  # type: List[ConfigField]
HTTP_AUTH_MAPPING = [
    CharConfigField(
        "auth.remote_user_header",
        "DF_REMOTE_USER_HEADER",
        help_str='Set it if the reverse-proxy authenticates users, a common value is "HTTP_REMOTE_USER". '
        "Note: the HTTP_ prefix is automatically added, just set REMOTE_USER in the "
        "reverse-proxy configuration. ",
    ),
    ListConfigField(
        "auth.remote_user_groups",
        "DF_DEFAULT_GROUPS",
        help_str="Comma-separated list of groups, for new users that are automatically created "
        "when authenticated by remote_user_header. Ignored if groups are read from a LDAP "
        "server. ",
    ),
]
LDAP_AUTH_MAPPING = [
    CharConfigField(
        "auth.ldap_server_url",
        "AUTH_LDAP_SERVER_URI",
        help_str='URL of your LDAP server, like "ldap://ldap.example.com". '
        'Python packages "pyldap" and "django-auth-ldap" must be installed.'
        "Can be used for retrieving attributes of users authenticated by the reverse proxy",
    ),
    CharConfigField(
        "auth.ldap_bind_dn",
        "AUTH_LDAP_BIND_DN",
        help_str="Bind dn for LDAP authentication",
    ),
    CharConfigField(
        "auth.ldap_bind_password",
        "AUTH_LDAP_BIND_PASSWORD",
        help_str="Bind password for LDAP authentication",
    ),
    CharConfigField(
        "auth.ldap_user_search_base",
        "AUTH_LDAP_USER_SEARCH_BASE",
        help_str="Search base for LDAP authentication by direct after an search, "
        'like "ou=users,dc=example,dc=com".',
    ),
    CharConfigField(
        "auth.ldap_filter",
        "AUTH_LDAP_FILTER",
        help_str='Filter for LDAP authentication, like "(uid=%%(user)s)" (the default),'
        ' the double "%%" is required in .ini files.',
    ),
    CharConfigField(
        "auth.ldap_direct_bind",
        "AUTH_LDAP_USER_DN_TEMPLATE",
        help_str="Set it for a direct LDAP bind and to skip the LDAP search, "
        'like "uid=%%(user)s,ou=users,dc=example,dc=com". '
        '%%(user)s is the only allowed variable and the double "%%" is required in .ini files.',
    ),
    BooleanConfigField(
        "auth.ldap_start_tls",
        "AUTH_LDAP_START_TLS",
        help_str='Set to "true" if you want to use StartTLS.',
    ),
    CharConfigField(
        "auth.ldap_first_name_attribute",
        "AUTH_LDAP_USER_FIRST_NAME",
        help_str='LDAP attribute for the user\'s first name, like "givenName".',
    ),
    CharConfigField(
        "auth.ldap_last_name_attribute",
        "AUTH_LDAP_USER_LAST_NAME",
        help_str='LDAP attribute for the user\'s last name, like "sn".',
    ),
    CharConfigField(
        "auth.ldap_email_attribute",
        "AUTH_LDAP_USER_EMAIL",
        help_str='LDAP attribute for the user\'s email, like "email".',
    ),
    CharConfigField(
        "auth.ldap_is_active_group",
        "AUTH_LDAP_USER_IS_ACTIVE",
        help_str='LDAP group DN for active users, like "cn=active,ou=groups,dc=example,dc=com"',
    ),
    CharConfigField(
        "auth.ldap_is_staff_group",
        "AUTH_LDAP_USER_IS_STAFF",
        help_str='LDAP group DN for staff users, like "cn=staff,ou=groups,dc=example,dc=com".',
    ),
    CharConfigField(
        "auth.ldap_is_superuser_group",
        "AUTH_LDAP_USER_IS_SUPERUSER",
        help_str='LDAP group DN for superusers, like "cn=superuser,ou=groups,dc=example,dc=com".',
    ),
    CharConfigField(
        "auth.ldap_require_group",
        "AUTH_LDAP_REQUIRE_GROUP",
        help_str="only authenticates users belonging to this group. Must be something like "
        '"cn=enabled,ou=groups,dc=example,dc=com".',
    ),
    CharConfigField(
        "auth.ldap_deny_group",
        "AUTH_LDAP_DENY_GROUP",
        help_str="authentication is denied for users belonging to this group. Must be something like "
        '"cn=disabled,ou=groups,dc=example,dc=com".',
    ),
    BooleanConfigField(
        "auth.ldap_mirror_groups",
        "AUTH_LDAP_MIRROR_GROUPS",
        help_str="Mirror LDAP groups at each user login",
    ),
    CharConfigField(
        "auth.ldap_group_search_base",
        "AUTH_LDAP_GROUP_SEARCH_BASE",
        help_str='Search base for LDAP groups, like "ou=groups,dc=example,dc=com"',
    ),
    ChoiceConfigFile(
        "auth.ldap_group_type",
        "AUTH_LDAP_GROUP_NAME",
        choices={
            "posix": "django_auth_ldap.config.PosixGroupType",
            "nis": "django_auth_ldap.config.NISGroupType",
            "GroupOfNames": "django_auth_ldap.config.GroupOfNamesType",
            "NestedGroupOfNames": "django_auth_ldap.config.NestedGroupOfNamesType",
            "GroupOfUniqueNames": "django_auth_ldap.config.GroupOfUniqueNamesType",
            "NestedGroupOfUniqueNames": "django_auth_ldap.config.NestedGroupOfUniqueNamesType",
            "ActiveDirectory": "django_auth_ldap.config.ActiveDirectoryGroupType",
            "NestedActiveDirectory": "django_auth_ldap.config.NestedActiveDirectoryGroupType",
            "OrganizationalRole": "django_auth_ldap.config.OrganizationalRoleGroupType",
            "NestedOrganizationalRole": "django_auth_ldap.config.NestedOrganizationalRoleGroupType",
        },
        help_str="Type of LDAP groups.",
    ),
]
LOG_MAPPING = [
    CharConfigField(
        "global.log_remote_url",
        "LOG_REMOTE_URL",
        help_str="Send logs to a syslog service. \n"
        "Examples: syslog+tcp://localhost:514/user, syslog:///local7 "
        "or syslog:///dev/log/daemon.",
        env_name="LOG_REMOTE_URL",
    ),
    CharConfigField(
        "global.log_sentry_dsn",
        "SENTRY_DSN",
        help_str="sentry DSN (see https://sentry.io/)",
        env_name="SENTRY_DSN",
    ),
    CharConfigField(
        "global.log_directory",
        "LOG_DIRECTORY",
        help_str="Write all local logs to this directory.",
        env_name="LOG_DIRECTORY",
    ),
    ChoiceConfigFile(
        "global.log_level",
        "LOG_LEVEL",
        help_str="Log level (one of 'debug', 'info', 'warn', 'error' or 'critical').",
        choices={
            "debug": "DEBUG",
            "info": "INFO",
            "warn": "WARN",
            "error": "ERROR",
            "critical": "CRITICAL",
        },
        env_name="LOG_LEVEL",
    ),
    BooleanConfigField(
        "global.log_remote_access",
        "LOG_REMOTE_ACCESS",
        help_str="If true, log of HTTP connections are also sent to syslog/logd.",
    ),
    FloatConfigField(
        "global.log_slow_query_duration_in_s",
        "LOG_SLOW_QUERY_DURATION_IN_S",
        help_str="Log slow queries that take more than this time (in seconds).",
        env_name="LOG_SLOW_QUERY_DURATION_IN_S",
    ),
]
PAM_AUTH_MAPPING = [
    BooleanConfigField(
        "auth.pam",
        "USE_PAM_AUTHENTICATION",
        help_str='Set to "true" if you want to activate PAM authentication',
    ),
]
RADIUS_AUTH_MAPPING = [
    CharConfigField(
        "auth.radius_server",
        "RADIUS_SERVER",
        help_str="IP or FQDN of the Radius server. "
        'Python package "django-radius" is required.',
    ),
    IntegerConfigField(
        "auth.radius_port", "RADIUS_PORT", help_str="port of the Radius server."
    ),
    CharConfigField(
        "auth.radius_secret",
        "RADIUS_SECRET",
        help_str="Shared secret if the Radius server",
    ),
]
REDIS_MAPPING = [
    CharConfigField(
        None,
        "COMMON_REDIS_URL",
        help_str="Redis database URL, for all redis things.",
        env_name="REDIS_URL",
    ),
    IntegerConfigField(
        "cache.db",
        "CACHE_DB",
        help_str="Database number (redis only). \n"
        'Python package "django-redis" is also required to use Redis.',
    ),
    CharConfigField(
        "cache.host", "CACHE_HOST", help_str="cache server host (redis or memcache)"
    ),
    CharConfigField(
        "cache.password",
        "CACHE_PASSWORD",
        help_str="cache server password (if required by redis)",
    ),
    IntegerConfigField(
        "cache.port", "CACHE_PORT", help_str="cache server port (redis or memcache)"
    ),
    ChoiceConfigFile(
        "cache.engine",
        "CACHE_PROTOCOL",
        choices={
            "redis": "redis",
            "memcache": "memcache",
            "locmem": "locmem",
            "file": "file",
        },
        help_str='cache storage engine ("locmem", "redis" or "memcache")',
    ),
    IntegerConfigField(
        "sessions.db",
        "SESSION_REDIS_DB",
        help_str="Database number of the Redis sessions DB\n"
        'Python package "django-redis-sessions" is required.',
    ),
    CharConfigField(
        "sessions.host", "SESSION_REDIS_HOST", help_str="Redis sessions DB host"
    ),
    CharConfigField(
        "sessions.password",
        "SESSION_REDIS_PASSWORD",
        help_str="Redis sessions DB password (if required)",
    ),
    IntegerConfigField(
        "sessions.port", "SESSION_REDIS_PORT", help_str="Redis sessions DB port"
    ),
]
SENDFILE_MAPPING = [
    BooleanConfigField(
        "global.use_apache",
        "USE_X_SEND_FILE",
        help_str='"true" if Apache is used as reverse-proxy with mod_xsendfile.'
        "The X-SENDFILE header must be allowed from file directories",
    ),
    ConfigField(
        "global.use_nginx",
        "X_ACCEL_REDIRECT",
        from_str=x_accel_converter,
        to_str=lambda x: "True" if x else "False",
        help_str='"true" is nginx is used as reverse-proxy with x-accel-redirect.'
        "The media directory (and url) must be allowed in the Nginx configuration.",
    ),
]
AUTH_MAPPING = (
    DJANGO_AUTH_MAPPING
    + RADIUS_AUTH_MAPPING
    + LDAP_AUTH_MAPPING
    + PAM_AUTH_MAPPING
    + HTTP_AUTH_MAPPING
)
# empty settings (please use the `social_authentications` management command instead)
INI_MAPPING = (
    ALLAUTH_MAPPING
    + AUTH_MAPPING
    + BASE_MAPPING
    + DATABASE_MAPPING
    + LOG_MAPPING
    + REDIS_MAPPING
    + SENDFILE_MAPPING
)
DEFAULT_INI_MAPPING = (
    BASE_MAPPING + DATABASE_MAPPING + LOG_MAPPING + REDIS_MAPPING + SENDFILE_MAPPING
)
EMPTY_INI_MAPPING = []



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
"""
universal server command

use gunicorn, uvicorn or daphne, depending on the selected options
"""
import sys

from django.conf import settings
from django.core.management import BaseCommand

from df_config.utils import is_package_present


class Command(BaseCommand):
    help = "Launch the server process"

    @property
    def listen_port(self):
        add, sep, port = settings.LISTEN_ADDRESS.partition(":")
        return int(port)

    @property
    def listen_address(self):
        add, sep, port = settings.LISTEN_ADDRESS.partition(":")
        return add

    def run_from_argv(self, argv):
        """
        Set up any environment changes requested (e.g., Python path
        and Django settings), then run this command. If the
        command raises a ``CommandError``, intercept it and print it sensibly
        to stderr. If the ``--traceback`` option is present or the raised
        ``Exception`` is not ``CommandError``, raise it.
        """
        if settings.DF_SERVER == "gunicorn":
            self.run_gunicorn()
        elif settings.DF_SERVER == "daphne":
            self.run_daphne()
        else:
            self.stderr.write(
                "unknown value '%s' for setting DF_SERVER. Please choose between 'daphne' and 'gunicorn'."
                % settings.DF_SERVER
            )
            return

    @staticmethod
    def get_wsgi_application():
        mod_name, sep, attr_name = settings.WSGI_APPLICATION.rpartition(".")
        return "%s:%s" % (mod_name, attr_name)

    @staticmethod
    def get_asgi_application():
        mod_name, sep, attr_name = settings.ASGI_APPLICATION.rpartition(".")
        return "%s:%s" % (mod_name, attr_name)

    def run_daphne(self):
        # noinspection PyPackageRequirements,PyUnresolvedReferences
        from daphne.cli import CommandLineInterface

        host, port = self.listen_address, self.listen_port
        app = self.get_asgi_application()

        class CLI(CommandLineInterface):
            def __init__(self):
                super().__init__()
                # noinspection PyProtectedMember
                for action in self.parser._actions:
                    if action.dest == "port":
                        action.default = port
                    elif action.dest == "host":
                        action.default = host
                    elif action.dest == "application":
                        action.default = app
                        action.required = False

        return CLI().run(sys.argv[2:])

    def run_gunicorn(self):
        sys.argv.pop(0)
        # noinspection PyPackageRequirements,PyUnresolvedReferences
        from gunicorn.config import KNOWN_SETTINGS, Setting

        # noinspection PyPackageRequirements,PyUnresolvedReferences
        from gunicorn.app.wsgiapp import WSGIApplication

        if settings.USE_WEBSOCKETS:
            application = self.get_asgi_application()
            worker_cls = "uvicorn.workers.UvicornWorker"
            if not is_package_present("uvicorn.workers"):
                self.stderr.write(
                    "you must install uvicorn to use websockets with Gunicorn."
                )
                return
        else:
            application = self.get_wsgi_application()
            worker_cls = "gunicorn.workers.gthread.ThreadWorker"

        class Application(WSGIApplication):
            def init(self, parser, opts, args):
                if not args:
                    args.append(application)
                super().init(parser, opts, args)

        for setting in KNOWN_SETTINGS:  # type: Setting
            if setting.name == "bind":
                setting.default = settings.LISTEN_ADDRESS
            elif setting.name == "worker_class":
                setting.default = worker_cls

        return Application("%(prog)s [OPTIONS] [APP_MODULE]").run()

    def handle(self, *args, **options):
        pass
